/*
** Copyright (C) Red Hat, Inc. 2007-2008
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "Network.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

counting_auto_ptr<Network::Hostent>
Network::getHostByName(const String& hostname)
{
	counting_auto_ptr<Hostent> ent_d(new Hostent());
	struct hostent *ent = 0;
	int error;

	gethostbyname2_r(hostname.c_str(),
		AF_INET,
		&(ent_d->ent),
		ent_d->data, sizeof(ent_d->data),
		&ent,
		&error);
	if (ent == &(ent_d->ent))
		return ent_d;
	throw String("unable to resolve ") + hostname;
}

std::vector<String>
Network::name2IP(const String& hostname)
{
	std::vector<String> addrs;

	try {
		char buff[INET_ADDRSTRLEN + 1];
		counting_auto_ptr<Hostent> hent = getHostByName(hostname);
		char **addrs_b = (*hent)->h_addr_list;

		for (int i = 0 ; addrs_b[i] ; i++) {
			struct in_addr addr;
			addr.s_addr = *((u_int32_t*) addrs_b[i]);
			if (inet_ntop(AF_INET, &addr, buff, sizeof(buff)))
				addrs.push_back(buff);
		}
	} catch ( ... ) {}
	return addrs;
}

String
Network::localhost()
{
	// get hostname
	char name[1024];
	if (gethostname(name, sizeof(name) - 1))
		return "";
	name[sizeof(name) - 1] = '\0';

	try {
		// get fqdn
		counting_auto_ptr<Hostent> ent = getHostByName(name);
		return String((*ent)->h_name);
	} catch ( ... ) {
		return name;
	}
}
