/*
** Copyright (C) Red Hat, Inc. 2005-2008
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "Server.h"
#include "ricci_defines.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>

extern "C" {
	void daemon_init(char *prog);
}

#include <iostream>


using namespace std;

bool foreground = false;
bool debug = false;
bool advertise_cluster = false;

int main(int argc, char** argv)
{
	uint32_t uid = 0;
	int32_t ricci_port = RICCI_SERVER_PORT;

	int rv;
	while ((rv = getopt(argc, argv, "cdfu:p:")) != EOF) {
		switch (rv) {
			case 'c':
				advertise_cluster = true;
				break;

			case 'd':
				debug = true;
				break;

			case 'f':
				foreground = true;
				break;

			case 'p':
				if (optarg != NULL) {
					uint32_t port;
					char *e = NULL;

					port = strtoul(optarg, &e, 10);
					if (*e != '\0' || !port || (port & 0xffff) != port) {
						fprintf(stderr, "Invalid port: %s\n", optarg);
						exit(1);
					}
					ricci_port = port;
				}
				break;

			case 'u':
				if (optarg != NULL) {
					char *e = NULL;
					uid = strtoul(optarg, &e, 10);
					if (*e != '\0') {
						fprintf(stderr, "Invalid uid: %s\n", optarg);
						exit(1);
					}
				}
				break;

			default:
				break;
		}
	}

	if (geteuid() != 0) {
		fprintf(stderr, "You must be root to run this program.\n");
		exit(1);
	}

	try {
		ServerSocket serv_sock(RICCI_SERVER_PORT);

		if (!foreground)
			daemon_init(argv[0]);

		if (uid != getuid()) {
			// change user
			if (setreuid(uid, uid)) {
				fprintf(stderr, "Error changing uid to %u: %s\n",
					uid, strerror(errno));
				exit(1);
			}
		}

		Server server(serv_sock);
		server.run();
	} catch ( String e ) {
		cerr << "exception: " << e << endl;
		exit(1);
	} catch ( ... ) {
		cerr << "exception caught" << endl;
		exit(1);
	}

	exit(0);
}
